#!/usr/bin/sage
from SageBase import SageBase

from sage.all import *


class EllipticCurveStandard(SageBase):

    INDENT = 40
    PLOT_OUTPUT_DIR = '../presentation'
    POINT_AT_INFINITY = (None, None)

    FONT_SIZE = 30
    POINT_SIZE = 75
    LINE_THICKNESS = 2
    PLOT_THICKNESS = 4

    POINT_COLOR = '#F57900'
    POINT_COLOR = '#CC0000'
    LINE_COLOR = POINT_COLOR
    PLOT_COLOR = '#486482'
    PLOT_COLOR_MINUS = PLOT_COLOR
#    PLOT_COLOR_MINUS = '#4E9A06' # comment this out to make them different
    LABEL_COLOR = POINT_COLOR
    LABEL_COLOR = '#000000'


    # the characteristic shape of the curve (x^3 + ax + b)
    # set these to None to accept command line parameters instead
    a = -5
    b = 5
    p = 0

    xMin = -2.6
    xMax = 4

    # this is a dict that remembers how many identical points have been registered,
    # and makes sure the labels don't step on each other.
    pointLabels = {}

    def doRunSteps(self):
        reload(sage.plot.plot) # caching
        if self.a == None:
            self.a = self.convertedArguments[0]
        if self.b == None:
            self.b = self.convertedArguments[1]

        if 4 * (self.a ** 3) + 27 * (self.b ** 2) == 0:
            raise "ERROR: Discriminant cannot be zero!"

        result = self.createEllipticCurve()
        if result:
            result = self.displayPointLabels(result)
            self.showPlot(result)


    # default implementation - change this in subclasses to draw the result
    def createEllipticCurve(self):
        return self.ellipticCurve()

    def ellipticCurve(self):
        self.out.put('plotting curve with a=%i, b=%i, p=%i' % (self.a, self.b, self.p))
        curve1 = plot(self.curveFunctionMinus,
            (self.xMin, self.xMax),
            rgbcolor=self.PLOT_COLOR_MINUS,
            thickness=self.PLOT_THICKNESS)
        curve2 = plot(self.curveFunction,
             (self.xMin, self.xMax),
             rgbcolor=self.PLOT_COLOR,
             thickness=self.PLOT_THICKNESS)
        result = curve1 + curve2
        return self.addLegend(result)

    def addLegend(self, result):
        result += text('a: %.1f, b: %.1f' % (self.a, self.b),
            (0.01, 0.01),
            vertical_alignment='bottom',
            horizontal_alignment='left',
            fontsize=self.FONT_SIZE,
            color=self.LABEL_COLOR,
            axis_coords=True)
        return result


    def curveFunction(self, x):
        result = (x ** 3) + (x * self.a) + self.b
        if result < 0:
            return None
        return sqrt(result)


    def curveFunctionMinus(self, x):
        result = self.curveFunction(x)
        if result == None:
            return None
        return result * -1


    def addPoints(self, p1, p2):
        self.out.indent("EllipticCurve.addPoints")
        self.printPoint("p1", p1)
        self.printPoint("p2", p2)

        if p1 == self.POINT_AT_INFINITY:
            self.printPoint("returning p2 because p1 is PAI", p2)
            self.out.unIndent()
            return p2
        if p2 == self.POINT_AT_INFINITY:
            self.printPoint("returning p1 because p2 is PAI", p1)
            self.out.unIndent()
            return p1

        #NOTE: for the modulo version, this may have points composed of Integer_mod objects
        (x1, y1) = p1
        (x2, y2) = p2

        if p1 == p2:
            # P == Q, so we double instead of add
            if y1 == 0:
                self.out.put("returning PAI because p1 == p2 at y=0")
                self.out.unIndent()
                return self.POINT_AT_INFINITY
            L = (3 * (x1 ** 2) + self.a) / (2 * y1)
        else:
            if x1 == x2:
                # we are adding P and -P.
                self.out.put("returning PAI because x1 == x2")
                self.out.unIndent()
                return self.POINT_AT_INFINITY
            L = (y2 - y1) / (x2 - x1)

        x3 = L ** 2 - x1 - x2
        y3 = L * (x1 - x3) - y1
        point = (x3, y3)
        self.printPoint("returning", point)
        self.out.unIndent()
        return point


    def doubleAndAdd(self, p, n):
        'return dP = P + P + ... P (n times)'
        # Do this by doubling for binary digit 0, double+add for binary digit 1
        binString = bin(n)[2:] # binary string representation of n
        self.out.indent('DoubleAndAdd(%s, %d) (binary value %s)' % (p, n, binString))
        q = p
        for i in range(1, len(binString)): # skip index 0 because it's always a 1
            self.out.indent('doubling 2x%s at index %i' % (q, i))
            q = self.addPoints(q, q)
            self.printPoint('result', q)
            self.out.unIndent()
            if binString[i] == '1':
                self.out.indent('adding %s+%s at index %i' % (q, p, i))
                q = self.addPoints(q, p)
                self.printPoint('result', q)
                self.out.unIndent()
            if q == self.POINT_AT_INFINITY:
                self.out.put('doubleAndAdd reached PAI at index %d for inputs %s, %i' % (i, p, n))
        self.out.unIndent()
        return q


    def point(self, p, label=None):
        if p == self.POINT_AT_INFINITY:
            p = (0, 0)
            if label:
                label = label + " (%s)" % u"\u221E"
        result = point(p, rgbcolor=self.POINT_COLOR, pointsize=self.POINT_SIZE, zorder=10)
        if label:
            x = p[0]
            y = p[1]
            if not x in self.pointLabels:
                self.pointLabels[x] = {}
            if not p[1] in self.pointLabels[x]:
                self.pointLabels[x][y] = []
            self.pointLabels[x][y].append(label)
        return result

    def displayPointLabels(self, result):
        minMaxData = result.get_minmax_data()
        padding = (minMaxData['xmax'] - minMaxData['xmin']) / 100
        lineHeightMultiplier = 0.035 * (minMaxData['ymax'] - minMaxData['ymin'])
        for x in self.pointLabels.keys():
            for y in self.pointLabels[x].keys():
                p = (float(x), float(y))
                labels = self.pointLabels[x][y]
                label = '\n'.join(labels)

                if len(labels) > 1:
                    # "join" the mulitple labels with a vertical line
                    vSize = len(labels) * lineHeightMultiplier
                    topPoint = (p[0] + padding, p[1] + vSize * 1.1)
                    bottomPoint = (p[0] + padding, p[1] - vSize * 0.9)
                    result += line([topPoint, bottomPoint], color=self.LABEL_COLOR)

                p = (p[0] + padding * 2, p[1])
                result += text(label,
                    p,
                    vertical_alignment='center',
                    horizontal_alignment='left',
                    fontsize=self.FONT_SIZE,
                    color=self.LABEL_COLOR)
        return result

    def finiteLine(self, p1, p2, lineStyle='-'):
        if p1 == self.POINT_AT_INFINITY:
            p1 = (0, 0)
        if p2 == self.POINT_AT_INFINITY:
            p2 = (0, 0)
        return line([p1, p2],
                    rgbcolor=self.LINE_COLOR,
                    thickness=self.LINE_THICKNESS,
                    linestyle=lineStyle)


    def showPlot(self, thePlot):
        fileName = '%s/%s' % (self.PLOT_OUTPUT_DIR, self.__class__.__name__)
        thePlot.save(fileName + '.png', figsize=[16, 9], legend_font_size='large')
        thePlot.save(fileName + '.pdf', figsize=[16, 9], legend_font_size='large')
#        thePlot.save(fileName + '.svg', figsize=[16, 9], legend_font_size='large')


class EllipticCurveModRing(EllipticCurveStandard):

    PLOT_COLOR_MINUS = '#4E9A06'

    p = 11

    xMin = -.5
    xMax = 12


    def doRunSteps(self):
        if mod(self.p, 4) != 3:
            raise "EllipticCurveModRing only works when p % 4 = 3!"
        return EllipticCurveStandard.doRunSteps(self)

    def addLegend(self, result):
        result += text('a: %i, b: %i, p: %i' % (self.a, self.b, self.p),
            (0.01, 0.01),
            vertical_alignment='bottom',
            horizontal_alignment='left',
            fontsize=self.FONT_SIZE,
            color=self.LABEL_COLOR,
            axis_coords=True)
        return result

    def curveFunction(self, x):
        x = mod(int(x), self.p)
        ySquared = x ** 3 + x * self.a + self.b
        y = ySquared ** ((self.p + 1) / 4)
        return y


    def addPoints(self, p1, p2):
        # force modulo arithmetic
        p1 = (mod(p1[0], self.p), mod(p1[1], self.p))
        p2 = (mod(p2[0], self.p), mod(p2[1], self.p))
        result = EllipticCurveStandard.addPoints(self, p1, p2)
        return result

    def swuHash(self, t):
        self.out.indent("SWU Hash Algorithm starting...")
        alpha = -(t ** 2)
        alphaDenominator = mod(alpha ** 2 + alpha, self.p)
        x2 = -(self.b / self.a) * (1 + 1 / alphaDenominator)
        h2 = mod(x2 ** 3 + self.a * x2 + self.b, self.p)
        exponent = (self.p + 1) / 4

        if self.legendre(h2) == 1:
            self.out.put("Found g(X2(t)) is a perfect square.")
            self.out.unIndent()
            return (x2, h2 ** exponent)

        self.out.put("Found g(X3(t)) is a perfect square.")
        x3 = alpha * x2
        h3 = x3 ** 3 + self.a * x3 + self.b
        self.out.unIndent()
        return (x3, h3 ** exponent)


    def legendre(self, x):
        # Legendre symbol:
        # 1 if x is Quadratic Residue mod p
        # 0 if x == 0 mod p
        # -1 otherwise
        exponent = ((self.p - 1) / 2)
        return x ** exponent

if __name__ == "__main__":
    # run ALL plots for the paper
    EllipticCurveStandard().run()
    EllipticCurveModRing().run()
