from datetime import datetime
from errno import EEXIST
from os import mkdir
from os.path import sep
import time
import sys

class JobOutput:

    INDENT_SPACER = '    '
    LOG_LEVEL_DISABLED = -1
    LOG_LEVEL_FATAL = 0
    LOG_LEVEL_ERROR = 1
    LOG_LEVEL_WARN = 2
    LOG_LEVEL_COMMON = 3
    LOG_LEVEL_VERBOSE = 4
    LOG_LEVEL_MUNDANE = 5
    LOG_LEVEL_DEBUG = 6
    TIME_FORMAT = '{0:%Y-%m-%d %H:%M:%S}'

    logFile = None
    logFileName = None
    logFileDir = None
    logFileThreshold = LOG_LEVEL_DISABLED
    displayOutputThreshold = LOG_LEVEL_COMMON
    indentStack = []
    messageQueue = []

    def initLog(self, logFileDir, logFileName):
        self.logFileThreshold = self.LOG_LEVEL_DEBUG
        if self.logFileName == None:
            self.logFileName = logFileName
        if self.logFileDir == None:
            self.logFileDir = logFileDir
        if self.logFile == None:
            # WARNING: these 2 steps don't use the System object, and therefore will not be mocked under --mock
            try:
                mkdir(self.logFileDir)
            except OSError, e:
                # Ignore directory exists error
                if e.errno != EEXIST:
                    raise
            self.logFile = open(self.getLogFileFullPath(), 'w')
        self.writeLogLine('== Initialized log ==')

    def getLogFileFullPath(self):
        logFileName = '%s%s%s.log' % (self.logFileDir, sep, self.logFileName)
        return logFileName

    def setJobOutputVerbosity(self, threshold):
        self.put("Setting display messaging threshold to %d" % threshold, self.LOG_LEVEL_MUNDANE)
        self.displayOutputThreshold = threshold

    def disableLogFile(self):
        self.logFileThreshold = self.LOG_LEVEL_DISABLED

    def disableJobOutput(self):
        self.displayOutputThreshold = self.LOG_LEVEL_DISABLED

    def put(self, message, logLevel=None):
        if logLevel == None:
            logLevel = self.LOG_LEVEL_COMMON
        self.addMessageToQueue(message, logLevel)
        self.flush()

    def addMessageToQueue(self, message, logLevel):
        messageArray = {'time':         time.time(),
                        'logLevel':     logLevel,
                        'indentLevel':  self.indentStack.__len__(),
                        'message':      message,
                        }
        self.messageQueue.append(messageArray)

    # WARNING!!! Not threadsafe! Consider rewriting. Should dequeue messages in a while() loop instead.
    def flush(self):
        for message in self.messageQueue[:]:
            self.flushSingleMessage(message)
        sys.stdout.flush()
        self.messageQueue = []

    def flushSingleMessage(self, message):
        printableText = (self.INDENT_SPACER * message['indentLevel']) + message['message']
        if message['logLevel'] <= self.logFileThreshold:
            self.writeLogLine(printableText)
        if message['logLevel'] <= self.displayOutputThreshold:
            print printableText

    def writeLogLine(self, line):
        timeStamp = self.TIME_FORMAT.format(datetime.today())
        line = '%s: %s\n' % (timeStamp, line.rstrip('\r\n'))
        self.logFile.write(line.encode('ascii', 'replace'))

    def indent(self, stackLabel):
        self.put(stackLabel)
        self.indentStack.append(stackLabel)
        self.flush

    def unIndent(self):
        try:
            stackLabel = self.indentStack.pop()
        except:
            # bug in calling code!!! 
            # Let the crappy output be the indicator of the bug, rather than an exception.
            pass
        # self.put(None, stackLabel)
        self.flush

    def start(self):
        pass

    def end(self):
        pass

if __name__ == '__main__':
    l = JobOutput()
    l.log('a test message', 'testMsg')
    l.log('a test message2', 'testMsg')
    l.log('a test message3', 'testMsg')
    l.log('a test message4', 'testMsg')
    l.log('a test message5', 'testMsg')
    l.indent('an indented level')
    l.log('a test message', 'testMsg')
    l.log('a test message2', 'testMsg')
    l.log('a test message3', 'testMsg')
    l.log('a test message4', 'testMsg')
    l.log('a test message5', 'testMsg')
    l.indent('an indented level')
    l.log('a test message', 'testMsg')
    l.log('a test message2', 'testMsg')
    l.log('a test message3', 'testMsg')
    l.log('a test message4', 'testMsg')
    l.log('a test message5', 'testMsg')
    l.unIndent()
    l.log('a test message', 'testMsg')
    l.log('a test message2', 'testMsg')
    l.log('a test message3', 'testMsg')
    l.unIndent()
    l.log('a test message', 'testMsg')
    l.log('a test message2', 'testMsg')
    l.log('a test message3', 'testMsg')
