#!/usr/bin/sage
from EllipticCurve import EllipticCurveModRing

from hashlib import sha1
from random import randint, seed

class SpekeSwu(EllipticCurveModRing):
    '''
    The actual implementation of the SPEKE protocol using NIST P-192 curve
    '''
    KNOWN_PASSPHRASE = "a cow jumped over the moon"

    # NIST Curve P-192 (ref: FIPS PUB 186-3)
    p = 6277101735386680763835789423207666416083908700390324961279
    n = 6277101735386680763835789423176059013767194773182842284081
    a = -3
    b = 0x64210519e59c80e70fa7e9ab72243049feb8deecc146b9b1
    Gx = 0x188da80eb03090f67cbf20eb43a18800f4ff0afd82ff1012
    Gy = 0x07192b95ffc8da78631011ed6b24cdd573f977a11e794811

    def createEllipticCurve(self):
        seed()
        self.out.indent("Alice and Bob are both generating h(pw) from passphrase: '%s'..." %
                         self.KNOWN_PASSPHRASE)

        sha1Hash = sha1()
        sha1Hash.update(self.KNOWN_PASSPHRASE)
        sha1HashInt = int(sha1Hash.hexdigest(), 16)
        self.printNumber("integer representation of sha1(pw)", sha1HashInt)

        self.h_pw = self.swuHash(sha1HashInt)
        self.printPoint("h(pw) = SWU(sha1(pw))", self.h_pw)
        self.out.unIndent()

        t = randint(0, 10000)
        self.out.indent("\nAlice is generating a private key using a random int %d..." % t)
        self.alice = self.swuHash(t)
        self.printPoint("Alice's private key", self.alice)
        self.a_h_pw = self.addPoints(self.alice, self.h_pw)
        self.printPoint("Alice generated public key A = a.h(pw)", self.a_h_pw)
        self.out.unIndent()

        t = randint(0, 10000)
        self.out.indent("\nBob is generating a private key using a random int %d..." % t)
        self.bob = self.swuHash(t)
        self.printPoint("Bob's private key", self.bob)
        self.b_h_pw = self.addPoints(self.bob, self.h_pw)
        self.printPoint("Bob generated public key B = b.h(pw)", self.b_h_pw)
        self.out.unIndent()

        self.out.indent("\nBob and Alice are exchanging keys...")
        self.ka = self.addPoints(self.alice, self.b_h_pw)
        self.printPoint("Alice generated Ka = a.(b.(h(pw)))", self.ka)

        self.kb = self.addPoints(self.bob, self.a_h_pw)
        self.printPoint("Bob generated Kb = b.(a.(h(pw)))", self.kb)
        self.out.unIndent()

        if self.ka[0] == self.kb[0] and self.ka[1] == self.kb[1]:
            self.out.put("Keys match! Alice and Bob now have a new shared secret!")
        else:
            self.out.put("ERROR: something went wrong.")

if __name__ == "__main__":
    SpekeSwu().run()
